# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


# ====================================================================
#
# eseas
#
from .cruncher_classes import Cruncher
from .folder_class import (
    DemetraFolder,
    JavaBinFolder,
    WorkspaceFolder,
    CruncherFolder,
)

# ====================================================================


class SingleOptions:
    """SingleOptions"""

    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super(SingleOptions, cls).__new__(cls)
        return cls.instance

    def set_items(cls, options):
        cls.instance.options = options


class SeasonalOptions:
    def __init__(
        self,
        demetra_folder=None,
        java_folder=None,
        local_folder=None,
        test=False,
        verbose=False,
        replace_original_files=False,
        auto_approve=False,
        result_file_names=(
            "sa",
            "s",
            "cal",
        ),
        workspace_mode=True,
        file_name_explanation=True,
        java_bin=None,
    ):

        self.demetra_folder = DemetraFolder(demetra_folder)
        self.java_folder = CruncherFolder(java_folder)
        self.local_folder = WorkspaceFolder(local_folder)
        self.java_bin = JavaBinFolder(java_bin)

        self.demetra_folder = str(self.demetra_folder)
        self.java_folder = str(self.java_folder)
        self.local_folder = str(self.local_folder)
        self.java_bin = java_bin if java_bin else None

        self.test = test
        self.verbose = verbose
        self.replace_original_files = replace_original_files
        self.auto_approve = auto_approve
        self.result_file_names = result_file_names
        self.workspace_mode = workspace_mode
        self.file_name_explanation = file_name_explanation

        self.set_options(workspace_mode)
        so = SingleOptions()
        so.set_items(self)

    def __repr__(self):
        template = f"""
    options= SeasonalOptions(
            demetra_folder="{self.demetra_folder}",
            java_folder="{self.java_folder}",
            local_folder="{self.local_folder}",
            test={self.test},
            verbose= {self.verbose},
            replace_original_files={self.replace_original_files},
            auto_approve={self.auto_approve},
            result_file_names={self.result_file_names},
            workspace_mode={self.workspace_mode},
            java_bin = {self.java_bin}
    )
        """
        return template

    def set_options(self, workspace_mode):
        c = Cruncher()
        c.set_items(
            local_work_space=self.local_folder,
            crunch_folder=self.java_folder,
            demetra_folder=self.demetra_folder,
            workspace_mode=workspace_mode,
            file_name_explanation=self.file_name_explanation,
            java_bin=self.java_bin,
        )



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


import pandas as pd
from io import StringIO
from typing import Tuple, Callable
from dataclasses import dataclass

if "main_" in __name__:
    from csv_content import template_csv, air_passenger
else:
    from .csv_content import template_csv, air_passenger


@dataclass
class eseas_DataFrame:
    name: str
    raw: str = None
    process_fnc: Callable = None
    df: pd.DataFrame = None

    def get_df(self) -> None:
        csv = StringIO(self.raw)
        df = pd.read_csv(csv, sep=";")
        self.df = df

    def process(self) -> None:
        self.get_df()

        if self.process_fnc is None:
            return
        self.df = self.process_fnc(self.df)

    def __str__(self) -> str:

        return f"<eseasDataFrame>{self.name}<name, df, raw>"

    def __repr__(self) -> str:
        return self.__str__()

    def to_df(self) -> pd.DataFrame:
        return self.df


def process_bal(df: pd.DataFrame) -> pd.DataFrame:
    if "Tarih" in df.columns:
        df = df.drop(["Unnamed: 4"], axis=1)
        df.set_index("Tarih", inplace=True)
    return df


def process_air(df: pd.DataFrame) -> pd.DataFrame:
    df["date"] = pd.to_datetime(df["Month"], format="%Y-%m")
    df.drop(["Month"], axis=1, inplace=True)
    df = df[["date", "#Passengers"]]

    return df


def get_sample_ts(template: str, name: str, fnc: Callable) -> eseas_DataFrame:

    e = eseas_DataFrame(name, template, fnc)
    e.process()
    return e


def get_sample_ts2(template: str, name: str) -> eseas_DataFrame:
    csv = StringIO(template)
    df = pd.read_csv(csv, sep=";")
    return eseas_DataFrame(name, df, template)


class SampleData:
    def get(self) -> list[eseas_DataFrame]:
        bal = get_sample_ts(template_csv, "bal", process_bal)
        air = get_sample_ts(air_passenger, "air", process_air)
        return [bal, air]


def get_sample_data() -> Tuple[eseas_DataFrame]:
    a = SampleData().get()
    # print(a)
    _ = tuple(map(check, a))
    return a


def check(e: eseas_DataFrame):
    print(e.name, "name ")
    print(e.df.columns.to_list(), "cols")
    print(e)
    assert isinstance(e, eseas_DataFrame)
    df = e.to_df()
    assert isinstance(df, pd.DataFrame)


if "main_" in __name__:

    a = SampleData().get()
    print(a)
    _ = tuple(map(check, a))



#!/usr/bin/env python
# coding: utf-8

import sys
import time
from pathlib import Path
import psutil
import subprocess

if sys.platform == "win32":
    import pythoncom
    import win32com.client


def get_input():
    ans = input(
        "In order to refresh this file \nEXCEL.EXE application will be stopped if it is running. \nConfirm ? [Y/n]"
    )
    return ans


def ask_permission() -> bool:
    ans = get_input()

    return ans.lower() in ["y", "yes", "ok", "good"]


def shut_excel(approve: bool = False):
    if not approve:
        answer = ask_permission()
        if not answer:
            raise ValueError("User did not confirm closing EXCEL.EXE application")

    command = ["cmd", "/c", "taskkill", "/f", "/im", "EXCEL.EXE"]
    print("closing Excel.exe application ...")
    try:
        result = subprocess.run(command, capture_output=True, text=True, check=True)
        print(result.stdout)
    except subprocess.CalledProcessError as e:
        print(f"Error: {e}")


excel_app = None


def excel_running():

    for proc in psutil.process_iter(["name"]):
        if proc.info["name"].lower() == "excel.exe":
            return True
    return False


class File:
    def __init__(self, name: str, folder: str = ".", approve: bool = False):
        self.name = name
        self.folder = folder
        self.path: Path = (self.folder / name).absolute()
        self.approve = approve

    def __str__(self):
        return f"<File [{self.path}]>"

    def exists(self):
        return self.path.exists()

    def check(self):
        if not self.exists():

            raise ValueError(f"File Not found ! {self.path}")
        assert self.exists()

    def refresh(self):
        global excel_app
        excel_app = get_excel_app(self.approve)

        try:
            print(f"working on: {self}")
            wb = excel_app.Workbooks.Open(self.path)
            wb.RefreshAll()
            time.sleep(1)
            wb.Save()
            wb.Close(False)
            print(f"Done => {self}")
        except Exception as e:
            import traceback

            traceback.print_exc()

            print(f"Error - {self}:\n{e}")


def check_and_get_files(files=None, root=".", approve: bool = False):

    fs = [File(x, root, approve) for x in files]
    for a in fs:
        a.check()

    for File_ in fs:
        yield File_


def get_excel_app(approve=False):
    global excel_app
    if excel_running():
        shut_excel(approve)

    excel_app = win32com.client.Dispatch("Excel.Application")
    return excel_app


def refresh(files, root: Path | str = ".", approve: bool = False):

    pythoncom.CoInitialize()

    for File_ in check_and_get_files(files, root, approve):
        File_.refresh()

    if excel_app is not None:
        try:
            excel_app.Quit()
            print("Before exiting closing excel application!")
        except Exception as quit_error:
            import traceback

            traceback.print_exc()
            print(f"Error while trying to close excel : {quit_error}")
    pythoncom.CoUninitialize()



from eseas import refresh, Path
from eseas.tests.test_utils import skip_if_unix

ROOT = Path(r"./excel_files")

@skip_if_unix
def test_refresh():
    if not ROOT.exists():
        return
    files = ["test.xlsx"]

    refresh(files, ROOT)



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


"""Demetra output types"""

from typing import Dict, Optional

full_content_dem = """
y|Original series
y_f|Forecasts of the original series
y_ef|Standard errors of the forecasts of the original
y_c|Interpolated series
yc_f|Forecasts of the interpolated series
yc_ef|Standard errors of the forecasts of the interpolated
y_lin|Linearised series (not transformed)
l|Linearised series (transformed)
ycal|Series corrected for calendar effects
ycal_f|Forecasts of the series corrected for calendar effects
l_f|Forecasts of the linearised series
l_b|Backcasts of the linearised series
t|Trend (including deterministic effects)
t_f|Forecasts of the trend
sa|Seasonally adjusted series (including deterministic effects)
sa_f|Forecasts of the seasonally adjusted series
s|Seasonal component (including deterministic effects)
s_f|Forecasts of the seasonal component
i|Irregular component (including deterministic effects)
i_f|Forecasts of the irregular component
det|All deterministic effects
det_f| Forecasts of the deterministic effects
cal|Calendar effects
cal_f| Forecasts of the calendar effects
tde|Trading day effect
tde_f|Forecasts of the trading day effect
mhe|Moving holidays effects
mhe_f|Forecasts of the moving holidays effects
ee|Easter effect
ee_f|Forecasts of the Easter effect
omhe|Other moving holidays effects
omhe_f|Forecasts of the other moving holidays effects
out|All outliers effects
out_f|Forecasts of all outliers effects
out_i|Outliers effects related to irregular (AO, TC)
out_i_f|Forecasts of outliers effects related to irregular (TC)
out_t|Outliers effects related to trend (LS)
out_t_f|Forecasts of outliers effects related to trend (LS)
out_s|Outliers effects related to seasonal (SO)
out_s_f|Forecasts of outliers effects related to seasonal (SO)
reg|All other regression effects
reg_f|Forecasts of all other regression effects
reg_i|Regression effects related to irregular
reg_i_f|Forecasts of regression effects related to irregular
reg_t|Regression effects related to trend
reg_t_f|Forecasts of regression effects related to trend
reg_s|Regression effects related to seasonal
reg_s_f|Forecasts of regression effects related to seasonal
reg_sa|Regression effects related to seasonally adjusted series
reg_sa_f|Forecasts of regression effects related to seasonally adjusted
reg_y|Separate regression effects
reg_y_f|Forecasts of separate regression effects
fullresiduals|Full residuals of the RegARIMA model
decomposition.y_lin|Linearised series used as input in the
decomposition.y_lin_f|Forecast of the linearised series used as
decomposition.t_lin|Trend produced by the decomposition
decomposition.t_lin_f|Forecasts of the trend produced by the
decomposition.s_lin|Seasonal component produced by the decomposition
decomposition.s_lin_f|Forecasts of the Seasonal component produced by
decomposition.i_lin|Irregular produced by the decomposition
decomposition.i_lin_f|Forecasts of the irregular produced by the
decomposition.sa_lin|Seasonally adjusted series produced by the decomposition
decomposition.sa_lin_f|Forecasts of the seasonally adjusted series produced
decomposition.si_lin|Seasonal-Irregular produced by the decomposition
lambda decomposition_a-tables_a1| For X-13ARIMA-SEATS only. Series
benchmarking_result| Benchmarked seasonally adjusted series
benchmarking_target| Target for the benchmarking
"""

local_separator = "|"

load_dict_once = False


def get_dict_of_demetra() -> Dict[str, str]:
    dem_dict = {}

    def make_dict(element):
        parts = element.split(local_separator)
        obj = {parts[0]: parts[1]}
        dem_dict.update(obj)
        return obj

    lines = full_content_dem.splitlines()
    lines = filter(lambda x: x.strip(), lines)
    _ = tuple(map(make_dict, lines))
    return dem_dict


if not load_dict_once:
    load_dict_once = get_dict_of_demetra()


def get_meaning_demetra_file(name: str) -> Optional[str]:
    global load_dict_once
    return load_dict_once.get(name, None)


def convert_demetra_file_md() -> str:
    items = list(load_dict_once.items())

    def f(item):
        k, v = item
        return f"{k} : {v } "

    return "\n".join(map(f, items))


def create_md_demetra() -> None:
    content = convert_demetra_file_md()
    from evdspy.EVDSlocal.common.files import Write

    Write("./demetra_components.md", content)


"""

    y : Original series
    y_f : Forecasts of the original series
    y_ef : Standard errors of the forecasts of the original
    y_c : Interpolated series
    yc_f : Forecasts of the interpolated series
    yc_ef : Standard errors of the forecasts of the interpolated
    y_lin : Linearised series (not transformed)
    l : Linearised series (transformed)
    ycal : Series corrected for calendar effects
    ycal_f : Forecasts of the series corrected for calendar effects
    l_f : Forecasts of the linearised series
    l_b : Backcasts of the linearised series
    t : Trend (including deterministic effects)
    t_f : Forecasts of the trend
    sa : Seasonally adjusted series (including deterministic effects)
    sa_f : Forecasts of the seasonally adjusted series
    s : Seasonal component (including deterministic effects)
    s_f : Forecasts of the seasonal component
    i : Irregular component (including deterministic effects)
    i_f : Forecasts of the irregular component
    det : All deterministic effects
    det_f :  Forecasts of the deterministic effects
    cal : Calendar effects
    cal_f :  Forecasts of the calendar effects
    tde : Trading day effect
    tde_f : Forecasts of the trading day effect
    mhe : Moving holidays effects
    mhe_f : Forecasts of the moving holidays effects
    ee : Easter effect
    ee_f : Forecasts of the Easter effect
    omhe : Other moving holidays effects
    omhe_f : Forecasts of the other moving holidays effects
    out : All outliers effects
    out_f : Forecasts of all outliers effects
    out_i : Outliers effects related to irregular (AO, TC)
    out_i_f : Forecasts of outliers effects related to irregular (TC)
    out_t : Outliers effects related to trend (LS)
    out_t_f : Forecasts of outliers effects related to trend (LS)
    out_s : Outliers effects related to seasonal (SO)
    out_s_f : Forecasts of outliers effects related to seasonal (SO)
    reg : All other regression effects
    reg_f : Forecasts of all other regression effects
    reg_i : Regression effects related to irregular
    reg_i_f : Forecasts of regression effects related to irregular
    reg_t : Regression effects related to trend
    reg_t_f : Forecasts of regression effects related to trend
    reg_s : Regression effects related to seasonal
    reg_s_f : Forecasts of regression effects related to seasonal
    reg_sa : Regression effects related to seasonally adjusted series
    reg_sa_f : Forecasts of regression effects related to seasonally adjusted
    reg_y : Separate regression effects
    reg_y_f : Forecasts of separate regression effects
    fullresiduals : Full residuals of the RegARIMA model
    decomposition.y_lin : Linearised series used as input in the
    decomposition.y_lin_f : Forecast of the linearised series used as
    decomposition.t_lin : Trend produced by the decomposition
    decomposition.t_lin_f : Forecasts of the trend produced by the
    decomposition.s_lin : Seasonal component produced by the decomposition
    decomposition.s_lin_f : Forecasts of the Seasonal component produced by
    decomposition.i_lin : Irregular produced by the decomposition
    decomposition.i_lin_f : Forecasts of the irregular produced by the
    decomposition.sa_lin : Seasonally adjusted series
      produced by the decomposition
    decomposition.sa_lin_f : Forecasts of the seasonally
      adjusted series produced
    decomposition.si_lin : Seasonal-Irregular produced by the decomposition
    lambda decomposition_a-tables_a1 :  For X-13ARIMA-SEATS only. Series
    benchmarking_result :  Benchmarked seasonally adjusted series
    benchmarking_target :  Target for the benchmarking

"""


if __name__ == "__main__":

    a = load_dict_once
    print(a)
    md = convert_demetra_file_md()
    print(md)



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


from pathlib import Path
from collections import OrderedDict
from typing import Optional
import os


def load_dotenv(filepath=".env") -> Optional[OrderedDict]:
    """Load environment variables from a .env file."""
    d = OrderedDict()

    if not os.path.exists(filepath):
        return

    with open(filepath, "r", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            key, sep, value = line.partition("=")
            if sep and key.strip():
                os.environ[key.strip()] = value.strip()
                d[key.strip()] = value.strip()
    return d


class FolderClass:
    type_: str = "FolderClass"

    def __init__(self, value: str = None):
        if value is None:
            self.value = None
        else:
            self.value = Path(value)
        self.check()

    def __repr__(self):
        return str(self.value)

    def __str__(self):
        return str(self.value)

    def warn_stop(self, msg: str = None):
        if msg:
            self.warn(msg, stop=True)
            return

        template = f"""
            {self.type_} does not exist
        """
        print(template)
        self.warn(template, stop=True)

    def warn(self, msg: str, stop=False):

        print(msg)
        self.sleep()
        if stop:
            raise ValueError(msg)

    def sleep(self, number: int = 2):
        import time

        time.sleep(number)

    def check_env_file(self):
        if self.env_file_exists():
            if self.value is None:
                v = self.get_from_env()
                if v is not None:
                    self.value = Path(v)

    def check(self):
        if self.env_file_exists():
            self.load_env()
            v = self.get_from_env()
            if v is not None:
                self.value = Path(v)
                return
        if not self.exists():
            self.warn_stop()

    def exists(self):
        if self.value is None:
            return False
        return Path(self.value).exists()

    def create(self):
        import os

        value = self.value
        print(f"creating folder {value}")
        if value is None:
            return False
        value = Path(value)

        try:
            os.makedirs(value, exist_ok=True)
        except OSError as e:
            print(f"Error creating folder {value} {e}")
            return False
        return True

    @staticmethod
    def load_env(env_file: str = ".env"):
        d = load_dotenv(env_file)
        if d is None:
            print(f" {env_file} not found. it will be ignored")
            return

        k = d.keys()
        if len(k) == 0:
            print(f"no keys found in {env_file}")
            return

    def env_file_exists(self, env_file: str = ".env"):
        return Path(env_file).exists

    def get_from_env(self):
        _ = self.load_env()
        if self.value is None:
            self.value = os.getenv(self.type_)
        return self.value


class TestFolderInstance(FolderClass):
    type_ = "TestFolderInstance"

    def check2(self):
        if not self.exists():
            self.warn(f"Test folder does not exist {self.value}")


class DemetraFolder(FolderClass):
    type_ = "demetra_source_folder"


class JavaBinFolder(FolderClass):
    type_ = "java_bin"

    def java_available(self):
        from eseas.core.java_environ import JavaEnviron

        print("Checking if java is available")
        j = JavaEnviron()
        p = j.find_java_paths()
        return bool(p)

    def check(self):
        self.check_env_file()

        if self.value is None:
            self.get_from_env()  # env is secondary

            if self.value is None:
                if self.java_available():
                    print(
                        """Java binary not given from the function or from env file but Java is available in the system."""
                    )
                    self.sleep(4)
                    return
                else:
                    self.warn_stop(f"Java binary folder does not exist {self.value}")

        if not self.exists():
            if self.java_available():
                print(
                    f"""Ignoring folder `{self.value}`since it does not exist and Java is available in the system."""
                )
                self.sleep(3)
                return
            self.warn_stop(f"Java binary folder does not exist {self.value}")


class CruncherFolder(FolderClass):
    type_ = "java_folder"


class WorkspaceFolder(FolderClass):
    type_ = "local_folder"

    def check(self):
        self.check_env_file()

        if not self.exists():
            self.warn("Creating folder for workspace since it does not exist yet.")
            if not self.create():
                self.warn_stop(f"Error creating folder {self.value}")



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


from dataclasses import dataclass
from pathlib import Path
import os
from eseas.core.utils_general2 import create_dir
from .utils_general2 import get_os


@dataclass
class TestingUtils:
    demetra_folder: str
    java_folder: str
    local_folder: str


def check_folder(folder):
    print(Path(folder).absolute())
    assert Path(folder).is_dir()


def get_env_java_folder():
    return os.environ.get("JAVA_CRUNCHER_BIN")


def get_testing_utils(check=False):
    """
    jdemetra/jswacruncher
    https://github.com/jdemetra/jwsacruncher/releases/tag/v2.2.4
    :return:
    """

    fold = "win"
    if get_os() != "win":
        fold = "unix"

    demetra_folder = rf"./eseas/data_for_testing/{fold}"
    # /Users/XABC/Downloads/jwsacruncher-2.2.5 3
    java_folder = Path(r"../../../Downloads/jwsacruncher-2.2.5 3/bin")

    if get_env_java_folder():
        java_folder = get_env_java_folder()

    # if GithubActions().is_testing():
    #     java_folder = os.environ["JAVA_CRUNCHER_BIN"]

    local_folder = r"./eseas_output"
    create_dir(local_folder)
    if check:
        _ = tuple(map(check_folder, (demetra_folder, java_folder, local_folder)))

    return TestingUtils(demetra_folder, java_folder, local_folder)



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


import os
import platform
from dataclasses import dataclass
import subprocess
from pathlib import Path



def check_java_version():
    JAVA_EXEC_PATH = "/usr/bin/java"
    os.system(f"{JAVA_EXEC_PATH} -version")


def get_os():
    return str(platform.system()).lower()



@dataclass
class JavaEnviron:

    def __init__(self, java_bin: str = None, quick=None):
        self.java_bin = None if java_bin is None else str(java_bin)
        self.check_exists()

        self.separator = ";" if self.windows() else ":"
        self.quick = True if quick is None else False
        if self.java_bin:
            self.quick = False

        self.process()

    def check_exists(self):
        if self.java_bin is None:
            return
        p = Path(self.java_bin)
        if p.exists():
            return
        raise ValueError(
            f"""\n\n `java_bin` directory was given as `{self.java_bin}` but it does not exist or it is not a directory `{self.java_bin}`
                         if your java folder is already defined as `path` just remove java_bin parameter while calling Cruncher. 
                         Or give a correct folder where java executable exists. 
                         """
        )

    def process(self):
        if self.java_bin is None:
            self.show_java_version()
            self.show_cmd()
            self.check_java_folders()
        else:
            self.set_path()

            self.sleep(3)
            self.show_java_version()
            self.show_cmd()
            self.check_java_folders()

    def windows(self) -> bool:
        return get_os() == "windows"

    def find_java_paths(self) -> list[str]:
        # cmd_where_which = "where" if self.windows() else "which"
        # result = self.run_process([cmd_where_which, "-a", "java"])
        cmd_list = ["where", "java"] if self.windows() else ["which", "-a", "java"]
        result = self.run_process(cmd_list)
        java_paths = result.stdout.strip().split("\n")
        return java_paths if java_paths[0] else []

    def run_process(self, cmds):
        class Result:
            stderr = "Error in subprocess while running " + " ".join(cmds)
            stdout = ""

        result = Result()
        try:
            result = subprocess.run(cmds, capture_output=True, text=True)
        except Exception:
            import traceback

            traceback.print_exc()

        return result

    def sleep(self, num=2):
        if self.quick:
            return
        import time

        time.sleep(num)

    def show_java_version(self):
        print("[showing java version] ...")
        result = self.run_process(["java", "-version"])
        res = (str(result.stdout) + str(result.stderr)).strip().split("\n")
        msg = "\n".join(res)
        print("Java Version \n", msg)
        self.sleep(2)

    def check_java_folders(self):
        print("[checking Java folders] ...")
        java_executables: list = self.find_java_paths()
        folders = self.split()
        active = False
        suffix = "/java.exe" if self.windows() else "/java"
        active_java_folder = "None"
        potential_java_folders = []
        for path in java_executables:
            p = str(path).removesuffix(suffix)
            if p in folders:
                if active is False:
                    active = True
                    active_java_folder = p
                    potential_java_folders.append((True, p))
                else:
                    potential_java_folders.append((False, p))

        folders = self.split()
        liste = []
        for p in folders:
            bool_ = active_java_folder == p
            liste.append((bool_, p))

        self.show_folders(
            potential_java_folders, "Current Folders in environment [Path variable]"
        )

        self.sleep(2)
        return potential_java_folders

    def show_cmd(self):
        cmd = """
        To check manually :
        from Terminal type  
                echo $PATH 
                java -version 

            """
        if self.windows():
            cmd = """
        To check manually :
        from Command line 
                path
                java -version 
            """
        print(cmd)
        self.sleep(2)

    def split(self) -> tuple:
        d = os.environ.get("PATH")
        return d.split(self.separator)

    def show_folders(self, items: tuple, msg: str = None):
        if not items:
            return
        if msg is None:
            msg = "Setting path environment variable with these folders:"

        def p(bool_: bool, string: str):
            string = str(string)
            if bool_:
                return "[+]  -> " + string
            return "     -> " + string

        def show():
            template = msg
            print(template)
            for bool_, item in items:
                x = p(bool_, item)
                if item is None:
                    continue
                print(x)

        show()
        self.sleep(2)

    def set_path(self):
        folders = self.split()
        folders2 = [self.java_bin] + [x for x in folders if x != self.java_bin]

        def check(string):
            return string == str(self.java_bin)

        items2 = [(check(x), x) for x in folders2]
        self.show_folders(items2)
        os.environ["PATH"] = self.separator.join(folders2)

        print(f"[Done] {self.java_bin} was set as java folder")
        self.sleep(2)


# e = JavaEnviron("/usr/bin")



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


from typing import Any
import pytest

import pandas as pd
from eseas import Seasonal
from eseas import Options
from eseas.core.df_operations import make_df_float
from eseas.core.cruncher_classes import get_cruncher
from eseas.core.cruncher_classes import Cruncher
from eseas.core.demetra import get_demetra_files
from eseas.core.picker_classes import OutFilePicker
from eseas.core.seas_testing_utils import get_testing_utils
from eseas.core.seas_utils import filter_xls
from eseas.core.df_operations import get_rand_hash
from eseas.tests.test_utils import skip_if_no_cruncher

# from .test_utils import skip_if_github

import random
from evdspy.EVDSlocal.common.file_classes import FileItem


def mock_file_items():
    items = []
    random.seed(954)
    exts = ("xlsx", "mat", "m", "xml")
    folders = (".", "A", "B", "C")
    for item in range(5):
        name = get_rand_hash(5)
        f = FileItem(random.choice(folders), name + "." + random.choice(exts))
        items.append(f)
    return items


testing_utils = get_testing_utils()
demetra_folder = testing_utils.demetra_folder
java_folder = testing_utils.java_folder
local_folder = testing_utils.local_folder


@skip_if_no_cruncher
def test_setup():
    options = Options(
        demetra_folder,
        java_folder,
        "ABC/A/B",  # it is ok if it does not exist
    )
    m = Seasonal(options)
    m.part1()
    m.part2()


# @skip_if_github
@skip_if_no_cruncher
def test_mevsimsel_general_basic():
    options = Options(
        demetra_folder,
        java_folder,
        local_folder,
    )
    m = Seasonal(options)
    m.part1()
    m.part2()


# @skip_if_github
@skip_if_no_cruncher
def test_c1():
    c = Cruncher()
    c.set_items(java_folder, local_folder, demetra_folder)
    dem_files = get_demetra_files(demetra_folder)
    for f in dem_files:
        of_picker = OutFilePicker(f)
        of_picker.pick_files()


# @skip_if_github
@skip_if_no_cruncher
def test_seasonal_general():
    options = Options(
        demetra_folder,
        java_folder,
        local_folder,
        test=False,
        verbose=False,
        replace_original_files=False,
        auto_approve=False,
        result_file_names=(
            "sa",
            "s",
            "cal",
        ),
        workspace_mode=True,
    )
    m = Seasonal(options)
    m.part1()
    m.part2()


@skip_if_no_cruncher
def test_seasonal_general_javabin_none():
    options = Options(
        demetra_folder,
        java_folder,
        local_folder,
        test=False,
        verbose=False,
        replace_original_files=False,
        auto_approve=False,
        result_file_names=(
            "sa",
            "s",
            "cal",
        ),
        workspace_mode=True,
        java_bin=None,
    )
    m = Seasonal(options)
    m.part1()
    m.part2()


@skip_if_no_cruncher
def test_seasonal_general_run():
    options = Options(
        demetra_folder,
        java_folder,
        local_folder,
        test=False,
        verbose=False,
        replace_original_files=False,
        auto_approve=False,
        result_file_names=(
            "sa",
            "s",
            "cal",
        ),
        workspace_mode=True,
        java_bin=None,
    )
    m = Seasonal(options)
    m.run(5)


@skip_if_no_cruncher
def test_seasonal_general_javabin_none_new_folder():
    options = Options(
        demetra_folder,
        java_folder,
        "new_folder",
        test=False,
        verbose=False,
        replace_original_files=False,
        auto_approve=False,
        result_file_names=(
            "sa",
            "s",
            "cal",
        ),
        workspace_mode=True,
        java_bin=None,
    )
    m = Seasonal(options)
    m.part1()
    m.part2()


@skip_if_no_cruncher
def test_seasonal_general_with_javabin():
    options = Options(
        demetra_folder,
        java_folder,
        local_folder,
        test=False,
        verbose=False,
        replace_original_files=False,
        auto_approve=False,
        result_file_names=(
            "sa",
            "s",
            "cal",
        ),
        workspace_mode=True,
        java_bin="/usr/bin",
    )
    m = Seasonal(options)
    m.part1()
    m.part2()



@skip_if_no_cruncher
def test_seasonal_general_with_wrong_javabin():
    with pytest.raises(ValueError):
        options = Options(
            demetra_folder,
            java_folder,
            local_folder,
            test=False,
            verbose=False,
            replace_original_files=False,
            auto_approve=False,
            result_file_names=(
                "sa",
                "s",
                "cal",
            ),
            workspace_mode=True,
            java_bin="/usr/something",
        )
        m = Seasonal(options)
        m.part1()
        m.part2()


def test_get_cruncher():
    c = Cruncher()
    c.set_items("", "", "")
    get_cruncher()


def test_make_df_float(capsys):
    with capsys.disabled():
        df = pd.DataFrame({"a": ["1,12", "12"], "b": [15, 20]})
        df2 = make_df_float(df)
        print(df2)


def test_filter_xls(capsys):
    with capsys.disabled():
        items = mock_file_items()
        a = filter_xls(items)
        print(a)


def test_Cruncher():
    c1 = Cruncher()
    c1.crunch_folder = "abc"
    c2 = Cruncher()
    c2.crunch_folder = "abcdefg"
    assert c1.crunch_folder == c2.crunch_folder


def test_make_float(capsys):
    data_dictA = {
        "s1": ["60,8456", "66,12656"],
        "s2": ["60,8456", "66,12656"],
    }
    data_dictB = {
        "s1": [60.8456, 66.12656],
        "s2": [60.8456, 66.12656],
    }
    data_dict = {
        # "donem": ["200201", "200202"],
        "donem": pd.Series([200201, 200202], dtype="int64"),
        "s1": ["60,8456", "66,12656"],
        "s2": ["60,8456", "66,12656"],
    }
    data_dict_target = {
        "donem": pd.Series([200201, 200202], dtype="int64"),
        "s1": [60.8456, 66.12656],
        "s2": [60.8456, 66.12656],
    }

    def func(
        _data_dict: dict[str, Any],
        _data_dict_target: dict[str, Any],
        except_columns=(),
        result=True,
    ):
        _df = pd.DataFrame.from_records(_data_dict)
        _df_target = pd.DataFrame.from_records(_data_dict_target)
        _new_df = make_df_float(_df, except_columns=except_columns)
        if result:
            assert _new_df.equals(_df_target)
        else:
            print(_new_df)
            print(_df_target)
            assert not _new_df.equals(_df_target)

    with capsys.disabled():

        func(data_dict, data_dict_target, except_columns=("donem",))
        func(data_dict, data_dict_target, except_columns=("BB",), result=False)

        func(data_dictA, data_dictB, except_columns=("AA",))
        func(data_dictA, data_dict_target, except_columns=("BB",), result=False)



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.

from .utils_general2 import create_dir
from .seas_utils import get_absolute

# ====================================================================
import traceback
from pathlib import Path
from rich import print
from evdspy.EVDSlocal.common.file_classes import make_eng
from evdspy.EVDSlocal.utils.utils_general import replace_recursive
import time
import tempfile
import os
from typing import Union


class Cruncher:
    """cruncher"""

    crunch_folder = "."
    local_work_space = "@wspace"
    demetra_folder: str = "."

    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super(Cruncher, cls).__new__(cls)
        return cls.instance

    def set_items(
        self,
        crunch_folder,
        local_work_space,
        demetra_folder,
        workspace_mode=True,
        file_name_explanation=True,
        java_bin=None,
    ):
        """Set Items Cruncher"""

        crunch_folder = str(crunch_folder)
        local_work_space = str(local_work_space)
        demetra_folder = str(demetra_folder)

        if java_bin:
            java_bin = Path(str(java_bin))
        self.java_bin = java_bin

        self.instance.crunch_folder = get_absolute(crunch_folder)
        self.instance.local_work_space = get_absolute(local_work_space)
        self.demetra_folder = get_absolute(demetra_folder)
        print("Options\n", "_" * 50, "\n")
        print("crunch_folder : ", crunch_folder)
        print("local_work_space : ", local_work_space)
        print("demetra_folder :", demetra_folder)
        print("\n", "_" * 50, "\n")

        # exit()

        control(
            self
        )  # it will raise if any of the folders does not exist except local workspace folder

        self.workspace_mode = workspace_mode
        self.file_name_explanation = file_name_explanation

        self.check_workspace_mode()
        self.set_environ_java(self.java_bin)

    def set_environ_java(self, java_bin: Path):
        from .java_environ import JavaEnviron

        _e = JavaEnviron(java_bin)

    def set_itemsObj(self, obj):
        """Set Items Cruncher"""

        self.instance.crunch_folder = get_absolute(obj.crunch_folder)
        self.instance.local_work_space = get_absolute(obj.local_work_space)
        self.demetra_folder = get_absolute(obj.demetra_folder)
        self.java_bin = get_absolute(obj.java_bin)
        control(self)
        self.workspace_mode = obj.workspace_mode
        self.file_name_explanation = obj.file_name_explanation
        self.check_workspace_mode()

    def create_workspace(self):
        def create_directory(address: Path):
            import os

            try:
                if not Path(address).is_dir():
                    os.makedirs(address)
                return True
            except Exception:
                traceback.print_exc()
                return False

        def path_str(Path_: Path):
            r = replace_recursive(str(Path_), "\\", ".")
            r = replace_recursive(r, ":", "..")
            return r

        def naming_format():
            p = Path() / self.instance.demetra_folder
            p = path_str(p)
            p = make_eng(p)
            return f"@eseas_wspace_{p}"

        if self.workspace_mode:
            ws = self.instance.local_work_space
            n_fname = Path() / ws / naming_format()
            if create_directory(n_fname):
                self.instance.local_work_space = n_fname

    def check_workspace_mode(self):
        # Cruncher
        # TODO SingleOptions().workspace_mode
        try:
            if self.workspace_mode:
                self.create_workspace()
        except Exception as exc:
            import traceback

            traceback.print_exc()
            raise exc


def show_cruncher_details_raise(cls):
    def check(folder: str):
        if Path(folder).is_dir():
            return "exists"
        return "does not exist!"

    msg = f"""\n\n
    _______________________________________________________________
    Could not find some folders.
    _______________________________________________________________
        java_folder     : [{check(cls.instance.crunch_folder)}]
                          {cls.instance.crunch_folder}
        local_workspace : [{check(cls.instance.local_work_space)}]
                          {cls.instance.local_work_space}
        demetra_folder  : [{check(cls.instance.demetra_folder)}]
                          {cls.instance.demetra_folder}
    """
    raise CruncherNotSet(msg)


def control(cls):
    if not check_cruncher(cls):
        show_cruncher_details_raise(cls)  # raise


def check_write_permission(folder: Union[str, Path]) -> bool:
    if not os.path.isdir(folder):
        raise NotADirectoryError(folder)
        # return False

    try:
        testfile = tempfile.TemporaryFile(dir=folder)
        testfile.close()
        return True
    except (OSError, IOError):
        return False


def check_read_permission(folder: Union[str, Path]) -> bool:
    if not os.path.isdir(folder):
        raise NotADirectoryError(folder)
        # return False
    return os.access(folder, os.R_OK)


def check_write_permission_create(folder):
    if not os.path.isdir(folder):
        create_dir(folder)
    return check_write_permission(folder)


def check_cruncher(cls) -> bool:
    msg_if_error = """
    Cruncher will need write permission to
    write permission
        - local workspace
        - java_folder
    read permission
        - demetra folder
    """

    a1 = check_write_permission(Path(cls.instance.crunch_folder))

    a2 = check_write_permission_create(Path(cls.instance.local_work_space))
    # this is for outputs therefore it will be created if it does not exist

    a3 = check_read_permission(Path(cls.instance.demetra_folder))

    ok = all(
        (
            a1,
            a2,
            a3,
        )
    )
    if not ok:
        print(msg_if_error)
        time.sleep(3)
        return False
        # raise CruncherNotSet()
    return ok


class CruncherNotSet(BaseException):
    pass


def get_cruncher():
    c = Cruncher()
    control(c)
    assert (
        c.crunch_folder is not None and c.local_work_space is not None
    ), "Cruncher not set!"
    return c


__all__ = [
    "get_cruncher",
    "Cruncher",
]



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.

"""
----------------------------------
      Global options
----------------------------------
"""

"""
    middle_folder
    .............
      -middle_folder : 'test_output'
      local_w_space / middle_folder / ...
      # produced csv content will be placed here
"""
middle_folder = "test_output"
"""
    demetra_command_file_name
    .............
       e.g. :  demetra_commands
"""
demetra_command_file_name = "demetra_commands"
"""
    max_num
    ...............
      this number will be maximum number of xml files that can be picked
      while searching for demetra xml files.
"""

max_num = 5_000



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


from rich import print
import pandas as pd
from evdspy.EVDSlocal.common.file_classes import file_items_update, FileItem
from .df_operations import make_df_float
from .seas_utils import get_xml_demetra, display

from ._options import demetra_command_file_name


def get_demetra_files(folder) -> list[FileItem]:
    xml_demetra = get_xml_demetra(folder)
    display(xml_demetra)
    xml_demetra = file_items_update(xml_demetra)  # [0:4]
    return xml_demetra


def convert_df_general(df: pd.DataFrame) -> pd.DataFrame:
    def ay_yil(yil, num, baslangic_ay):
        def ay_hesapla():
            ay = (num + baslangic_ay) % int(period)
            if ay == 0:
                return 12
            return ay

        def yil_hesapla():
            artik = 0
            if num + baslangic_ay > 12:
                artik = 1
            return yil + artik

        sonuc = yil_hesapla() * 100 + ay_hesapla()
        return sonuc

    df.index = range(len(df))
    df_new = df.copy()
    df_new.drop([0, 1, 2, 3, 4], axis=0, inplace=True)
    period = list(df.iloc[1])[0]
    baslangic_yil = list(df.iloc[2])[0]
    baslangic_ay = list(df.iloc[3])[0]
    adet = list(df.iloc[4])[0]
    adet = int(adet)
    liste = tuple(ay_yil(baslangic_yil, x, baslangic_ay) for x in range(adet))
    df_new.columns = df.iloc[0]
    df_new["donem"] = liste
    df_new.set_index("donem", drop=True, inplace=True)
    df_new = make_df_float(df_new)
    return df_new


def show_title(msg: str) -> None:
    print("=" * 50)
    print(" " * 5, msg)
    print("=" * 50)


def write_bat_file_demetra(xml_demetra, file_name=demetra_command_file_name) -> None:
    from .create_bat_command import (
        create_command,
        copy_xml_files_local,
        copy_folder_demetra,
        write_bat_file,
    )

    cmds = create_command(xml_demetra)
    copy_xml_files_local(xml_demetra)
    copy_folder_demetra(xml_demetra)
    text = "".join(cmds)
    show_title(f"Creating command file to call `jwsacruncher` : [{file_name}]")

    write_bat_file(text, file_name)



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


from dataclasses import dataclass
from pathlib import Path
import time
from evdspy.EVDSlocal.common.colors import print_with_failure_style
from evdspy.EVDSlocal.utils.utils_general import create_directory
from .create_bat_command import run_bat_commands
from .cruncher_classes import Cruncher
from .demetra import get_demetra_files, write_bat_file_demetra
from .picker_classes import OutFilePicker
from .seas_utils import view_display
from evdspy.EVDSlocal.common.files import Write
from .seasonal_adv_utils import (
    get_input_from_user,
    common_space_msg,
    seasonal_results_msg,
    get_results_info,
    display,
)

from ._options import demetra_command_file_name

# demetra_command_file_name = 'demetra_commands'


class SeasonalADV:
    def __init__(self, options):
        self.options = options
        self.code_reproduce()

    def sleep(self, seconds: int = 2):
        print(f"Sleeping for {seconds} seconds.")
        time.sleep(seconds)

    def part1(self):
        self.common_space_check()

    def part2(self):
        self.seasonal_results_advanced()

    def run(self, seconds: int = 10):
        """Run part1 and part2 with a sleep in between"""
        self.part1()
        self.sleep(seconds)
        self.part2()

    def common_space_check(self):
        """common_space_check"""
        msg = common_space_msg(self.options.test, self.options.demetra_folder)
        view_display(msg)
        time.sleep(1)
        xml_demetra = get_demetra_files(self.options.demetra_folder)
        if self.options.test:
            xml_demetra = xml_demetra[0:20]
        fn = demetra_command_file_name
        write_bat_file_demetra(xml_demetra, file_name=fn)
        if not self.options.auto_approve and not get_input_from_user():
            view_display(
                "demetra command did not run.\n"
                "You may type y next time if you like them to run."
            )
            return
        run_bat_commands()

    def seasonal_results_advanced(self):
        """seasonal_results_advanced"""
        view_display(seasonal_results_msg(self.options.test))
        time.sleep(1)
        if self.options.replace_original_files:
            t = get_results_info(Cruncher().local_work_space)
            print(t)
        xml_demetra = get_demetra_files(self.options.demetra_folder)
        if self.options.test:
            xml_demetra = xml_demetra[0:5]
        if self.options.verbose:
            display(xml_demetra)
        for file in xml_demetra:
            of_picker = OutFilePicker(
                file,
                names=self.options.result_file_names,
                file_name_explanation=self.options.file_name_explanation,
            )
            of_picker.pick_files()
        return xml_demetra

    def code_reproduce(self):
        return ReproduceMevsimsel(self).code_reproduce()


@dataclass
class ReproduceMevsimsel:
    def __init__(self, mevsimsel: SeasonalADV):
        self.mevsimsel = mevsimsel

    def code_reproduce(self):
        template = self.code_reproduce_template()
        folder_name = Path() / Cruncher().local_work_space / "@codes_reproduce"
        create_directory(folder_name)
        file_name = folder_name / "code_run.py"
        try:
            Write(file_name, template)
        except Exception as exc:
            print(exc)
            print_with_failure_style(
                f"Could not create reproduce file for later use.  {str(exc)}"
            )

    def code_reproduce_template(self):
        from datetime import datetime

        date_str = datetime.now()
        template = f"""\n
#======================================================================
# This script was created by eseas    {date_str}
#   In order to reproduce you may use this template
#======================================================================
from eseas import SeasonalOptions
from eseas import Seasonal
def main():
    {self.mevsimsel.options.__repr__()}
    m = Seasonal(options)
    m.part1()
    m.part2()
main()
        """
        return template


__all__ = ["SeasonalADV"]



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


import traceback
from abc import ABC
from pathlib import Path
import pandas as pd
from evdspy.EVDSlocal.common.file_classes import FileItem
from ._options import middle_folder
from .cruncher_classes import get_cruncher
from .demetra import convert_df_general
from .demetra_file_naming import get_meaning_demetra_file

"""
FilePicker classes are designed to harvest
seasonal output data from CSV files created by the cruncher
and convert them into Pandas DataFrames.

"""


class FilePicker(ABC):
    def __init__(
        self,
        file_item,
        names=(
            "sa",
            "s",
            "cal",
        ),
    ):
        self.file_item = file_item
        self.names = names


class OutFilePicker(FilePicker):
    def __init__(
        self,
        file_item,
        names=(
            "sa",
            "s",
            "cal",
        ),
        file_name_explanation=True,
    ):
        super().__init__(file_item, names)
        self.file_name_explanationOption = file_name_explanation

    def process_file(self, file_name, name) -> None:
        df = pd.read_csv(
            file_name,
            engine="python",
            encoding="unicode_escape",
            sep=";",
            index_col=False,
            header=None,
        )
        df = df.T
        df = convert_df_general(df)
        e = self.file_name_explanationOption
        fname = get_name_format(self.file_item, name, get_explanation=e)
        df.to_excel(fname)

    def pick_files(self):
        for name in self.names:
            try:
                self.process_file(get_file_name_x(self.file_item, name), name)
            except Exception:
                traceback.print_exc()
                msg = f"output [{name}] not found in JDemetra output!"
                print(msg)


def get_file_name_x(x: FileItem, name="sa") -> Path:
    """while reading output files"""
    explanation = get_explanation_if_neces(name, False)
    # middle_folder: 'test_output'
    return (
        Path()
        / get_cruncher().local_work_space
        / middle_folder
        / x.encoded_name
        / "SAProcessing-1"
        / f"series_{name}{explanation}.csv"
    )


def get_explanation_if_neces(name_type, get_explanation=False) -> str:
    if not get_explanation:
        return ""
    name_exp = get_meaning_demetra_file(name_type)
    if not name_exp:
        name_exp = "explanation_na"
    name_exp = "_{" + name_exp + "}"

    return name_exp


def get_name_format(file_item: FileItem, name_type="sa", get_explanation=True):
    """While writing output files

    Naming these files is optional as they do not
    necessarily need to be read by the script again.

    """
    explanation = get_explanation_if_neces(name_type, get_explanation)
    if not isinstance(file_item, FileItem):
        return Path("None.xlsx")
    return Path(get_cruncher().local_work_space) / str(
        file_item.short_name + f"_{name_type}{explanation}_eseas_" + ".xlsx"
    )



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


from eseas.core.seasonal_general import SeasonalADV as Seasonal
from eseas.core.seasonal_general import SeasonalADV
from eseas.core.seasonal_options import SeasonalOptions as Options
from eseas.core.seasonal_options import SeasonalOptions
from eseas.core.cruncher_classes import Cruncher
from eseas.data_for_testing.some_data import get_sample_data
from eseas.core.utils_general2 import walk, walk2, sleep
from pathlib import Path
import sys

if sys.platform == "win32":
    from eseas.core.refresh_excel import refresh
else:
    refresh = None

__all__ = [
    "Cruncher",
    "SeasonalADV",
    "SeasonalOptions",
    "Seasonal",
    "Options",
    "get_sample_data",
    "walk",
    "walk2",
    "Path" ,
    "refresh",
    "sleep"
    
]



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


from typing import Tuple, Optional

import pandas as pd
from pathlib import Path


def get_rand_hash(num=5) -> str:
    import secrets

    return secrets.token_urlsafe(nbytes=num)


def try_to_write_excel(df: pd.DataFrame, file_name: Path) -> None:
    try:
        df.to_excel(file_name)
    except Exception:
        pa = file_name.parent
        hash_ = get_rand_hash(5)
        file_name_new = file_name.stem + f"_{hash_}.xlsx"
        print(Path(pa) / file_name_new)
        df.to_excel(Path(pa) / file_name_new)


def rreplace(st: str, e, y) -> str:
    if e not in st:
        return st
    st = st.replace(e, y)
    return rreplace(st, e, y)


def sayi_donustur(pot_sayi: str) -> float:
    if not pot_sayi:
        return pot_sayi
    try:
        if isinstance(pot_sayi, float):
            return pot_sayi
        if str(pot_sayi).isnumeric():
            return float(pot_sayi)
        pot_sayi = str(pot_sayi)
        pot_sayi = rreplace(pot_sayi, ".", "")
        pot_sayi = pot_sayi.replace(",", ".")
        if not pot_sayi.replace(".", "").isnumeric():
            return pot_sayi
        return float(pot_sayi)
    except Exception:
        return pot_sayi


def convert_df_number(df: pd.DataFrame, except_columns: Tuple[str] = ()):
    def convert_number_item(item):
        try:
            new_number = sayi_donustur(item)  # float(str(item).replace(",", "."))
        except Exception as exc:
            print(item)
            print(exc)
            new_number = 0

        return new_number

    def convert_numbers(numbers):
        return tuple(map(convert_number_item, numbers))

    if except_columns is None:
        except_columns = tuple()
    for column in df.columns:
        if column in except_columns:
            continue
        df[column] = convert_numbers(list(df[column]))

    return df


def make_df_float(df: pd.DataFrame, except_columns: Optional[Tuple[str]] = None):
    if except_columns is None:
        except_columns = ("donem",)
    df = convert_df_number(df, except_columns)
    return df



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.



from eseas import Seasonal
from eseas import Options
from eseas.core.seas_testing_utils import get_testing_utils
from eseas.tests.test_utils import skip_if_no_cruncher


def create_env_file():
    testing_utils = get_testing_utils()
    demetra_folder = testing_utils.demetra_folder
    java_folder = testing_utils.java_folder
    local_folder = testing_utils.local_folder

    template = f"""
java_bin = /usr/bin
java_folder = {java_folder}
demetra_source_folder = {demetra_folder}
local_folder = {local_folder}

"""
    with open(".env", "w") as f:
        f.write(template)


create_env_file()


@skip_if_no_cruncher
def test_seasonal_environment_file():
    options = Options(
        None,
        None,
        None,
        test=False,
        verbose=False,
        replace_original_files=False,
        auto_approve=False,
        result_file_names=(
            "sa",
            "s",
            "cal",
        ),
        workspace_mode=True,
        java_bin=None,
    )
    m = Seasonal(options)
    m.part1()
    m.part2()


@skip_if_no_cruncher
def test_seasonal_environment_file2():
    options = Options(
        demetra_folder=None,
        java_folder=None,
        local_folder=None,
        test=False,
        verbose=False,
        replace_original_files=False,
        auto_approve=False,
        result_file_names=(
            "sa",
            "s",
            "cal",
        ),
        workspace_mode=True,
        java_bin=None,
    )
    m = Seasonal(options)
    m.part1()
    m.part2()


@skip_if_no_cruncher
def test_with_empty_options():
    from eseas import Seasonal, Options
    import time

    # Load options from the `.env` file
    options = Options()

    # Initialize and execute the seasonal adjustment process
    m = Seasonal(options)
    m.part1()
    time.sleep(2)  # Pause before running part2
    m.part2()



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


import functools
import re
from .utils_general2 import list_apply, tuple_apply, dict_apply

strategies = {"strict": r"\W+", "light": r"^[a-zA-Z0-9_ ]", "medium": r"\W+ "}


def base_replace(string, level="strict"):
    bound_func = functools.update_wrapper(
        functools.partial(base_replace, level=level), base_replace
    )
    if isinstance(string, (list,)):
        return list_apply(string, bound_func)
    if isinstance(string, (tuple,)):
        return tuple_apply(string, bound_func)
    if isinstance(string, (dict,)):
        return dict_apply(string, bound_func)
    reg_strategy = strategies[level]
    return re.sub(reg_strategy, "_", string)


def replace_strict(string):
    return base_replace(string, "strict")


def replace_light(string):
    return base_replace(string, "light")


def replace_medium(string):
    return base_replace(string, "medium")



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


from eseas.core.github_actions import GithubActions
import pytest
import sys
import os


def gth_testing():
    return GithubActions().is_testing()


reason_gth = "passing when github Actions "


skip_if_github = pytest.mark.skipif(gth_testing(), reason=reason_gth)


def has_cruncher():
    """Check if jwsacruncher is available via JAVA_CRUNCHER_BIN environment variable."""
    cruncher_bin = os.environ.get("JAVA_CRUNCHER_BIN")
    return cruncher_bin is not None and cruncher_bin.strip() != ""


# Use custom marker - the pytest_runtest_setup hook in conftest.py will handle skipping
skip_if_no_cruncher = pytest.mark.skip_if_no_cruncher


def is_unix():
    """Check if the system is Unix-based (macOS or Linux)."""
    return sys.platform in ("darwin", "linux")


skip_if_unix = pytest.mark.skipif(
    is_unix(),
    reason="Skipping on Unix systems (macOS/Linux) - Windows-only features",
)



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


import json
from pathlib import Path
import typing as t


def dict_apply(dict_: dict, func: t.Callable, args=None):
    if isinstance(dict_, (tuple, list)):
        return list(map(func, dict_))
    new_dict_ = {}
    for k, v in dict_.items():
        if isinstance(v, str):
            obj = {k: func(v)}
            new_dict_.update(obj)
    return new_dict_


def remove_json_bad_chars(json_object_str):
    if isinstance(json_object_str, (tuple, list)):
        return list(map(remove_json_bad_chars, json_object_str))
    if isinstance(json_object_str, (dict)):
        return dict_apply(json_object_str, remove_json_bad_chars)
    bad_chars = ("$", "%", "&", "?", "Ä", "ÄŸ")
    replace_chars = ("USD", "PERCENT", "AMPERSTAND", "QUESTIONMARK", "G", "g")
    for index, bad_char in enumerate(bad_chars):
        new_char = replace_chars[index]
        json_object_str = replace_recursive_json(json_object_str, bad_char, new_char)
    return json_object_str


def replace_recursive_json(text: str, char: str, new_char: str):
    parts = text.split(char)
    return new_char.join(parts)


def create_json_file(json_dict: dict, file_name="file_and_cols.js"):
    json_object_str: str = json.dumps(json_dict, ensure_ascii=False)
    json_object_str = "file_and_cols='" + json_object_str + "'"
    json_file_name = Path() / file_name
    from evdspy.EVDSlocal.common.files import Write

    print("writing....", json_file_name)
    result, msg = Write(Path() / json_file_name, json_object_str)
    print(result, msg)
    result, msg = Write(Path(r"some_folder/json") / json_file_name, json_object_str)
    print(result, msg)



# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.

import os
import shutil
import subprocess
import urllib.request
import zipfile
from pathlib import Path
import pytest
import time


CRUNCHER_VERSION = "2.2.5"
CRUNCHER_URL = f"https://github.com/jdemetra/jwsacruncher/releases/download/v{CRUNCHER_VERSION}/jwsacruncher-{CRUNCHER_VERSION}-bin.zip"
CRUNCHER_EXTRACT_PATH = Path(__file__).parent.parent.parent / "jwsacruncher"
CRUNCHER_BIN_PATH = CRUNCHER_EXTRACT_PATH / f"jwsacruncher-{CRUNCHER_VERSION}" / "bin"
MAX_RETRIES = 3
RETRY_DELAY = 2  # seconds


def download_jwsacruncher():
    """Download jwsacruncher if not already present, with retry logic."""
    if CRUNCHER_BIN_PATH.exists():
        print(f"jwsacruncher already exists at {CRUNCHER_BIN_PATH}")
        return True

    print(f"\nDownloading jwsacruncher {CRUNCHER_VERSION}...")
    CRUNCHER_EXTRACT_PATH.mkdir(parents=True, exist_ok=True)

    zip_file = CRUNCHER_EXTRACT_PATH / f"jwsacruncher-{CRUNCHER_VERSION}-bin.zip"

    for attempt in range(MAX_RETRIES):
        try:
            print(f"Download attempt {attempt + 1}/{MAX_RETRIES}...")
            
            # Create a custom request with a User-Agent header
            req = urllib.request.Request(
                CRUNCHER_URL,
                headers={'User-Agent': 'Mozilla/5.0'}
            )
            
            with urllib.request.urlopen(req, timeout=60) as response:
                with open(zip_file, 'wb') as out_file:
                    out_file.write(response.read())
            
            print(f"Downloaded to {zip_file}")

            print(f"Extracting to {CRUNCHER_EXTRACT_PATH}...")
            with zipfile.ZipFile(zip_file, "r") as zip_ref:
                zip_ref.extractall(CRUNCHER_EXTRACT_PATH)

            zip_file.unlink()  # Remove zip file after extraction
            print(f"jwsacruncher extracted successfully to {CRUNCHER_BIN_PATH}")
            return True

        except Exception as e:
            print(f"Download attempt {attempt + 1} failed: {e}")
            if zip_file.exists():
                zip_file.unlink()
            
            if attempt < MAX_RETRIES - 1:
                print(f"Waiting {RETRY_DELAY} seconds before retry...")
                time.sleep(RETRY_DELAY)
            else:
                print(f"Failed to download jwsacruncher after {MAX_RETRIES} attempts")
                return False


@pytest.fixture(scope="session", autouse=True)
def setup_cruncher():
    """Automatically download and set up jwsacruncher for tests."""
    # Check if JAVA_CRUNCHER_BIN is already set
    if os.environ.get("JAVA_CRUNCHER_BIN"):
        cruncher_bin = os.environ.get("JAVA_CRUNCHER_BIN")
        print(f"Using existing JAVA_CRUNCHER_BIN: {cruncher_bin}")
        
        # Verify it exists
        if not Path(cruncher_bin).exists():
            print(f"Warning: JAVA_CRUNCHER_BIN points to non-existent path: {cruncher_bin}")
        return

    # Try to download
    if download_jwsacruncher():
        # Set the environment variable
        os.environ["JAVA_CRUNCHER_BIN"] = str(CRUNCHER_BIN_PATH)
        print(f"Set JAVA_CRUNCHER_BIN to {os.environ['JAVA_CRUNCHER_BIN']}")
    else:
        print("WARNING: Could not download jwsacruncher. Tests requiring the cruncher will be skipped.")


def pytest_configure(config):
    """Hook that runs before test collection to setup cruncher and register markers."""
    # Ensure download happens before any test collection
    if not os.environ.get("JAVA_CRUNCHER_BIN"):
        if download_jwsacruncher():
            os.environ["JAVA_CRUNCHER_BIN"] = str(CRUNCHER_BIN_PATH)
            print(f"Set JAVA_CRUNCHER_BIN to {os.environ['JAVA_CRUNCHER_BIN']}")
        else:
            print("WARNING: Could not download jwsacruncher. Tests requiring the cruncher will be skipped.")
    
    # Register custom markers
    config.addinivalue_line(
        "markers", "skip_if_no_cruncher: skip test if jwsacruncher is not available"
    )


def pytest_runtest_setup(item):
    """Skip tests at runtime if they require cruncher but it's not available."""
    if item.get_closest_marker("skip_if_no_cruncher"):
        if not os.environ.get("JAVA_CRUNCHER_BIN"):
            pytest.skip("jwsacruncher not available (JAVA_CRUNCHER_BIN not set)")




# This file is part of the eseas project
# Copyright (C) 2024 Sermet Pekin
#
# This source code is free software; you can redistribute it and/or
# modify it under the terms of the European Union Public License
# (EUPL), Version 1.2, as published by the European Commission.
#
# You should have received a copy of the EUPL version 1.2 along with this
# program. If not, you can obtain it at:
# <https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12>.
#
# This source code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# European Union Public License for more details.
#
# Alternatively, if agreed upon, you may use this code under any later
# version of the EUPL published by the European Commission.


import typing as t
from pathlib import Path
import os
import random
from typing import Union, Iterable


from evdspy.EVDSlocal.common.file_classes import FileItem
from evdspy.EVDSlocal.common.colors import (
    print_excel_created_style,
    print_with_info_style,
    print_with_updating_style,
    print_with_success_style,
)

from ._options import max_num

color_funcs = (
    print_excel_created_style,
    print_with_info_style,
    print_with_updating_style,
    print_with_success_style,
    print,
)


def get_print_color() -> callable:
    return random.choice(color_funcs)


def display_line(line):
    indent = "" * 15
    result = f"{indent}{line}"
    return result


def display_lines(lines):
    st: str = "\n".join(tuple(map(display_line, lines)))
    func = get_print_color()
    func(st)


def view_display(*msg):
    for item in msg:
        item = str(item)
        display_lines(item.splitlines())


class SingleOptions:
    """SingleOptions"""

    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super(SingleOptions, cls).__new__(cls)
        return cls.instance

    def set_items(cls, options):
        cls.instance.options = options


def get_absolute(folder: str):
    p = Path(folder)
    return Path(p.absolute())


def clean_file_name(file_name: str):
    file_name = file_name.replace("\n", " ")
    file_name = file_name.replace(":", " ")
    file_name = file_name.replace("'", " ")
    file_name = file_name.replace("$", " ")
    return file_name.replace("\\", "..__")


def filter_xls(items: t.List[FileItem]):
    def check(item):
        def check_this(ext):
            return str(item).endswith(ext)

        return any(
            map(
                check_this,
                (
                    "xlsx",
                    "xlsm",
                    "xls",
                ),
            )
        )

    return tuple(x for x in items if check(x.full_name))


def filter_xml(items: t.List[FileItem]):
    return tuple(x for x in items if str(x.full_name).endswith("xml"))


def list_of_folders(folder: Union[str, Path]) -> list[str]:
    import os

    def check(d):
        return os.path.isdir(os.path.join(folder, d))

    return [d for d in os.listdir(folder) if check(d)]


def filter_xml_demetra(items: t.List[FileItem]):
    def stem(file_item: FileItem):
        return str(file_item.full_name.stem)

    def check_demetra(file_item: FileItem):
        """
        SomeFolder_JDemtra.xml
        Calendars.xml
        SAProcessing-1.xml
        Vars-1.xml
        """

        def check_if_folder_exists() -> bool:
            folders = list_of_folders(file_item.dir_path)
            fname = file_item.full_name
            if not hasattr(file_item, "short_name"):
                if str(fname).endswith(".xml"):
                    file_item.short_name = fname.stem
                else:

                    return False

            return file_item.short_name in folders

        def check_all_parts(part_func: t.Callable):
            def check_this_part(item: str):
                part: str = part_func(file_item)
                result = item.strip().lower() in part.strip().lower()
                return result

            return any(
                map(
                    check_this_part,
                    (
                        "jdemtra",
                        "demetra",
                        "seas_adj",
                    ),
                )
            )

        return any(map(check_all_parts, (stem,))) or check_if_folder_exists()

    items = filter_xml(items)

    return list(x for x in items if check_demetra(x))


class MaxFileNumberReached(BaseException): ...


def list_files_recursive(folder: t.Union[str, Path]) -> list[FileItem]:
    res = []
    break_loop = False
    if not Path(folder).is_dir():
        raise FileNotFoundError(folder)
    try:
        for dir_path, dir_names, file_names in os.walk(folder):
            print(f"{folder}>> {dir_path}   ")
            for file_name in file_names:
                if break_loop:
                    break
                file_item = FileItem(dir_path, file_name)
                if max_num < len(file_item.created_items):
                    view_display(
                        "limit max number reached  you "
                        " may change it from search_main.py file "
                    )
                    n = len(file_item.created_items)
                    raise MaxFileNumberReached(f"Max{n}")
                res.append(file_item)
    except MaxFileNumberReached:
        print("max number reached... returning limit number of files...")
    return res


def display(some_files: Iterable[FileItem], max_num=10):
    template = ""
    msg = f"""
=====================================================
        Number of files found :  {len(some_files)}
=====================================================
    """
    template += msg + chr(10)
    for index, item in enumerate(some_files):
        if index < max_num:
            template += item.file_name + chr(10)
    view_display(template)


def search_demetra_folder(
    root: t.Union[str, Path] = None,
    filter_func: callable = None,
) -> list[FileItem]:
    if not root:
        raise ValueError
    f = list_files_recursive(root)
    if not f or not callable(filter_func):
        return f

    files_filtered = filter_func(f)
    return files_filtered


def get_type_of_files_demetra(folder, filter_func=filter_xls):
    files = search_demetra_folder(folder, filter_func)
    return files


def get_type_of_files(folder, filter_func=filter_xls):
    files = search_demetra_folder(folder, filter_func)  # search_folders()
    return files


def get_xml_demetra(folder) -> list[FileItem]:
    return get_type_of_files_demetra(folder, filter_xml_demetra)
